(*----------------------------------------------------------------------*)
 
IMPLEMENTATION MODULE M_BitOps;

FROM Portab IMPORT tCompiler, Compiler;
(*$?Compiler=Haenisch:
(* IMPLEMENTATION FR  >>> Hnisch-Modula-2 <<< *)
(*                                              *)
(*$V-   Range-Checks                            *)
(*$T-   Parameter-Checks                        *)
(*$S-   Stack-Check                             *)
(*$Y-   Runtime-Checks                          *)
(*$I-   keine Variablen-Initialisierung         *)
(*                                              *)
(*----------------------------------------------*)
 *)

(*$?Compiler=Megamax:
(* IMPLEMENTATION FR  >>> Megamax-Modula-2 <<< *)
(*                                              *)
(*$R-   Range-Checks                            *)
(*$S-   Stack-Check                             *)
(*                                              *)
(*----------------------------------------------*)
 *)


IMPORT SYSTEM, M_Types;
FROM SYSTEM IMPORT LOC;

TYPE    tByte =         POINTER TO M_Types.ByteSet;
        tWord =         POINTER TO BITSET;
        tLong =         POINTER TO M_Types.LONGBITSET;

VAR     b1, b2, b3:     tByte;
        w1, w2, w3:     tWord;
        l1, l2, l3:     tLong;

VAR     b:              POINTER TO SYSTEM.BYTE;
        w:              POINTER TO SHORTCARD;
        l:              POINTER TO LONGCARD;

VAR     cast:   RECORD
                 CASE : SHORTINT OF
                  0: a: SYSTEM.ADDRESS;|
                  1: li: LONGCARD;|
                  2: h,l: SHORTCARD;|
                  3: b1, b2, b3, b4: SYSTEM.LOC;|
                 END;
                END;

PROCEDURE BitOp (op: Operation; c1, c2: ARRAY OF LOC; VAR val: ARRAY OF LOC);
BEGIN
 IF (HIGH (c1) = HIGH (c2)) AND (HIGH (c1) = HIGH (val)) THEN
  CASE HIGH (val) OF
   0: b1:= SYSTEM.ADR (c1);
      b2:= SYSTEM.ADR (c2);
      b3:= SYSTEM.ADR (val);
      CASE op OF
       and: b3^:= b1^ * b2^;|
       or:  b3^:= b1^ + b2^;|
       xor: b3^:= b1^ / b2^;|
      END;
      |
   1: w1:= SYSTEM.ADR (c1);
      w2:= SYSTEM.ADR (c2);
      w3:= SYSTEM.ADR (val);
      CASE op OF
       and: w3^:= w1^ * w2^;|
       or:  w3^:= w1^ + w2^;|
       xor: w3^:= w1^ / w2^;|
      END;
      |
   3: l1:= SYSTEM.ADR (c1);
      l2:= SYSTEM.ADR (c2);
      l3:= SYSTEM.ADR (val);
      CASE op OF
       and: l3^:= l1^ * l2^;|
       or:  l3^:= l1^ + l2^;|
       xor: l3^:= l1^ / l2^;|
      END;
      |
   ELSE ;
  END; (* CASE *)
 END; (* IF *)
END BitOp;

PROCEDURE SetBit (pos: SHORTCARD; VAR val: ARRAY OF LOC);
BEGIN
 CASE HIGH (val) OF
  0: b1:= SYSTEM.ADR (val);
     IF pos <  8 THEN INCL (b1^, pos);  END;
     |
  1: w1:= SYSTEM.ADR (val);
     IF pos < 16 THEN INCL (w1^, pos);  END;
     |
  3: l1:= SYSTEM.ADR (val);
     IF pos < 32 THEN INCL (l1^, pos);  END;
     |
  ELSE ;
 END; (* CASE *)
END SetBit;

PROCEDURE ClearBit (pos: SHORTCARD; VAR val: ARRAY OF LOC);
BEGIN
 CASE HIGH (val) OF
  0: b1:= SYSTEM.ADR (val);
     IF pos <  8 THEN EXCL (b1^, pos);  END;
     |
  1: w1:= SYSTEM.ADR (val);
     IF pos < 16 THEN EXCL (w1^, pos);  END;
     |
  3: l1:= SYSTEM.ADR (val);
     IF pos < 32 THEN EXCL (l1^, pos);  END;
     |
  ELSE ;
 END; (* CASE *)
END ClearBit;

PROCEDURE IsSet (pos: SHORTCARD; val: ARRAY OF LOC): BOOLEAN;
BEGIN
 CASE HIGH (val) OF
  0: b1:= SYSTEM.ADR (val);
     IF pos <  8 THEN  RETURN pos IN b1^;  END;
     |
  1: w1:= SYSTEM.ADR (val);
     IF pos < 16 THEN  RETURN pos IN w1^;  END;
     |
  3: l1:= SYSTEM.ADR (val);
     IF pos < 32 THEN  RETURN pos IN l1^;  END;
     |
  ELSE ;
 END; (* CASE *)
END IsSet;

PROCEDURE ShiftLeft  (VAR val: ARRAY OF LOC; num: SHORTCARD);
VAR w: SHORTCARD;
    l: LONGCARD;
BEGIN
 CASE HIGH (val) OF
  0: cast.li:= 0;  cast.b4:= val[0];  w:= cast.l;
     IF num <  8 THEN
      WHILE num > 0 DO  EXCL (b1^, 7);  w:= w * 2;  DEC (num);  END;
     END;
     |
  1: cast.li:= 0;  cast.b3:= val[0];  cast.b4:= val[1];  w:= cast.l;
     IF num < 16 THEN
      WHILE num > 0 DO  EXCL (w1^, 15);  w:= w * 2;  DEC (num);  END;
     END;
     |
  3: cast.b1:= val[0]; cast.b2:= val[1]; cast.b3:= val[2]; cast.b4:= val[3];
     l:= cast.li;
     IF num < 32 THEN
      WHILE num > 0 DO  EXCL (l1^, 31);  l:= l * 2;  DEC (num);  END;
     END;
     |
  ELSE ;
 END; (* CASE *)
END ShiftLeft;

PROCEDURE ShiftRight (VAR val: ARRAY OF LOC; num: SHORTCARD);
BEGIN
 CASE HIGH (val) OF
  0: w:= SYSTEM.ADR (val);
     IF num <  8 THEN
      WHILE num > 0 DO  w^:= w^ DIV 2;  DEC (num);  END;
     END;
     |
  1: w:= SYSTEM.ADR (val);
     IF num < 16 THEN
      WHILE num > 0 DO  w^:= w^ DIV 2;  DEC (num);  END;
     END;
     |
  3: l:= SYSTEM.ADR (val);
     IF num < 32 THEN
      WHILE num > 0 DO  l^:= l^ DIV 2;  DEC (num);  END;
     END;
     |
  ELSE ;
 END; (* CASE *)
END ShiftRight;

END M_BitOps.



