IMPLEMENTATION MODULE M_Buffer;

FROM Portab IMPORT tCompiler, Compiler;
(*$?Compiler=Haenisch:
(* IMPLEMENTATION FR  >>> Hnisch-Modula-2 <<< *)
(*                                              *)
(*$V-   Range-Checks                            *)
(*$T-   Parameter-Checks                        *)
(*$S-   Stack-Check                             *)
(*$Y-   Runtime-Checks                          *)
(*$I-   keine Variablen-Initialisierung         *)
(*                                              *)
(*----------------------------------------------*)
 *)

(*$?Compiler=Megamax:
(* IMPLEMENTATION FR  >>> Megamax-Modula-2 <<< *)
(*                                              *)
(*$R-   Range-Checks                            *)
(*$S-   Stack-Check                             *)
(*                                              *)
(*----------------------------------------------*)
 *)

FROM Storage    IMPORT ALLOCATE, DEALLOCATE;
FROM SYSTEM     IMPORT  ADDRESS, ADR, TSIZE, LOC;

CONST   cMax =          07FFFH;

TYPE    INFO =          POINTER TO ARRAY [0..cMax] OF LOC;

TYPE    ENTRY =         POINTER TO Entry;
        Entry =         RECORD
                         addr: INFO;
                         size: SHORTINT;
                         next: ENTRY;
                        END;

TYPE    BUFFER =        POINTER TO Buffer;
        Buffer =        RECORD
                         start: ENTRY;
                         end:   ENTRY;
                         entry: LONGINT;
                        END;

PROCEDURE Copy (from, to: INFO; size: SHORTINT);
VAR c: SHORTINT;
BEGIN
 FOR c:= 0 TO size DO to^[c]:= from^[c]; END;
END Copy;

PROCEDURE NewBuffer (VAR buffer: BUFFER): BOOLEAN;
BEGIN
 ALLOCATE (buffer, TSIZE(Buffer));  
 IF buffer = NIL THEN RETURN FALSE; END;
 buffer^.start:= NIL;  
 buffer^.end:= NIL;
 buffer^.entry:= LONGCARD (0);
 RETURN TRUE;
END NewBuffer;

PROCEDURE DisposeBuffer (VAR buffer: BUFFER);
VAR p: ENTRY;
BEGIN
 IF buffer # NIL THEN
  WITH buffer^ DO
   WHILE start # NIL DO
    p:= start^.next;
    DEALLOCATE (start^.addr, 0);  
    DEALLOCATE (start, 0);  
    start:= p;
   END;
  END;
  DEALLOCATE (buffer, 0);  
 END;
END DisposeBuffer;

PROCEDURE BufferEmpty (buffer: BUFFER): BOOLEAN;
BEGIN
 IF buffer = NIL THEN  RETURN FALSE;  END;
 RETURN buffer^.start = NIL;
END BufferEmpty;

PROCEDURE BufferEntries (buffer: BUFFER): LONGINT;
BEGIN
 IF buffer = NIL THEN  RETURN LONGCARD (0);
                 ELSE  RETURN buffer^.entry;
 END;
END BufferEntries;

PROCEDURE Put (buffer: BUFFER; infoptr: ADDRESS; size: SHORTINT): BOOLEAN;
VAR p: ENTRY;
    i: INFO;
BEGIN
 IF buffer = NIL THEN  RETURN FALSE;  END;
 ALLOCATE (p, TSIZE(Entry));  
 IF p = NIL THEN  RETURN FALSE;  END;
 p^.size:= size;
 p^.next:= NIL;
 i:= infoptr;
 ALLOCATE (p^.addr, VAL(LONGCARD, p^.size));
 IF p^.addr = NIL THEN  DEALLOCATE (p, 0);   RETURN FALSE;  END;
 Copy (i, p^.addr, p^.size);
 WITH buffer^ DO
  IF end # NIL THEN
   end^.next:= p;  end:= p;
  ELSIF start # NIL THEN
   start^.next:= p;  start:= p;
  ELSE
   start:= p;  end:= p;
  END;
  INC (entry);
 END;
 RETURN TRUE;
END Put;

PROCEDURE Get (buffer: BUFFER; infoptr: ADDRESS): BOOLEAN;
VAR p: ENTRY;
    i: INFO;
BEGIN
 IF buffer = NIL THEN  RETURN FALSE;  END;
 WITH buffer^ DO
  IF start = NIL THEN  RETURN FALSE;  END;
  p:= start;
  IF infoptr # NIL THEN  RETURN FALSE;  END;
  i:= infoptr;
  Copy (p^.addr, i, p^.size);
  start:= p^.next;
  DEALLOCATE (p^.addr, 0);  
  DEALLOCATE (p, 0);  
  IF start = NIL THEN  end:= NIL;  END;
  DEC (entry);
 END;
 RETURN TRUE;
END Get;

END M_Buffer.

