IMPLEMENTATION MODULE M_Lists;

FROM Portab IMPORT tCompiler, Compiler;
(*$?Compiler=Haenisch:
(* IMPLEMENTATION FR  >>> Hnisch-Modula-2 <<< *)
(*                                              *)
(*$V-   Range-Checks                            *)
(*$T-   Parameter-Checks                        *)
(*$S-   Stack-Check                             *)
(*$Y-   Runtime-Checks                          *)
(*$I-   keine Variablen-Initialisierung         *)
(*                                              *)
(*----------------------------------------------*)
 *)

(*$?Compiler=Megamax:
(* IMPLEMENTATION FR  >>> Megamax-Modula-2 <<< *)
(*                                              *)
(*$R-   Range-Checks                            *)
(*$S-   Stack-Check                             *)
(*                                              *)
(*----------------------------------------------*)
 *)

FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM SYSTEM IMPORT  ADDRESS, ADR, TSIZE, LOC;

CONST   cMax =          07FFFH;

TYPE    INFO =          POINTER TO ARRAY [0..cMax] OF LOC;

TYPE    ENTRY =         POINTER TO Entry;
        Entry =         RECORD
                         addr: INFO;
                         size: SHORTINT;
                         next: ENTRY;
                         last: ENTRY;
                        END;

TYPE    LIST =          POINTER TO List;
        List =          RECORD
                         start: ENTRY;
                         end:   ENTRY;
                         comp:  LCompProc;
                         entry: LONGINT;
                        END;


PROCEDURE Copy (from, to: INFO; size: SHORTINT);
VAR c: SHORTINT;
BEGIN
 FOR c:= 0 TO size DO to^[c]:= from^[c]; END;
END Copy;

PROCEDURE NewList (VAR list: LIST; comp: LCompProc): BOOLEAN;
BEGIN
 ALLOCATE (list,  TSIZE (List));  
 IF list = NIL THEN RETURN FALSE; END;
 list^.start:= NIL;
 list^.comp:= comp;
 RETURN TRUE;
END NewList;

PROCEDURE DisposeList (VAR list: LIST);
VAR p: ENTRY;
BEGIN
 IF list # NIL THEN
  WITH list^ DO
   WHILE start # NIL DO
    p:= start^.next;
    DEALLOCATE (start^.addr, 0);  
    DEALLOCATE (start, 0);  
    start:= p;
   END;
  END;
  DEALLOCATE (list, 0);  
 END;
END DisposeList;

PROCEDURE NilLEntry (): ENTRY;
BEGIN
 RETURN NIL;
END NilLEntry;

PROCEDURE InsertLEntry (list: LIST; info: ADDRESS; size: SHORTINT): BOOLEAN;
VAR s, t: ENTRY;
    i:    INFO;
    f:    BOOLEAN;
BEGIN
 IF list = NIL THEN  RETURN FALSE;  END;
 ALLOCATE (t, TSIZE(Entry));
 IF t = NIL THEN  RETURN FALSE;  END;
 t^.size:= size;
 t^.last:= NIL;
 t^.next:= NIL;
 ALLOCATE (t^.addr, VAL(LONGCARD, t^.size));
 IF t^.addr = NIL THEN  DEALLOCATE (t, 0);    RETURN FALSE;  END;
 i:= info;
 Copy (i, t^.addr, t^.size);
 WITH list^ DO
  s:= start;
  IF s = NIL THEN
   start:= t;  end:= t;
  ELSE
   f:= FALSE;
   WHILE (s # NIL) AND NOT f DO
    IF comp (s^.addr, i) = smaller THEN s:= s^.next; ELSE  f:= TRUE; END;
   END;
   t^.next:= s;
   IF s = start THEN
    start:= t;  s^.last:= t
   ELSIF s = NIL THEN
    t^.last:= end;  end^.next:= t;  end:= t;
   ELSE
    t^.last:= s^.last;  s^.last^.next:= t;  s^.last:= t;
   END;
  END;
  INC (entry);
 END;
 RETURN TRUE;
END InsertLEntry;

PROCEDURE ListEntries (list: LIST): LONGINT;
BEGIN
 IF list = NIL THEN RETURN LONGCARD (0);
               ELSE RETURN list^.entry;
 END;
END ListEntries; 

PROCEDURE SearchLEntry (list: LIST; from: ENTRY;
                       info: ADDRESS;  key: LCompProc): ENTRY;
VAR s: ENTRY;
BEGIN
 IF list = NIL THEN  RETURN NIL;  END;
 WITH list^ DO
  s:= start;
  WHILE s # NIL DO
   IF key (s^.addr, info) = equal THEN  RETURN s;  END;
   s:= s^.next;
  END;
 END;
 RETURN NIL;
END SearchLEntry;

PROCEDURE DeleteLEntry (list: LIST; VAR entr: ENTRY);
VAR t: ENTRY;
BEGIN
 IF (list = NIL) OR (entr = NIL) THEN  RETURN;  END;
 WITH list^ DO
  IF entr = start THEN
   t:= start;  start:= start^.next;  start^.last:= NIL;
   IF start = NIL THEN  end:= NIL;  END;
   DEALLOCATE (t^.addr, 0);    DEALLOCATE (t, 0);  
   DEC (entry);
  ELSIF entr = end THEN
   t:= end;  end:= end^.last;  end^.next:= NIL;
   DEALLOCATE (t^.addr, 0);    DEALLOCATE (t, 0);  
   DEC (entry);
  ELSE
   entr^.next^.last:= entr^.last;
   entr^.last^.next:= entr^.next;
   DEALLOCATE (entr^.addr, 0); DEALLOCATE (entr, 0);
   DEC (entry);
  END;
 END;
 entr:= NIL;
END DeleteLEntry;

PROCEDURE FirstLEntry (list: LIST): ENTRY;
BEGIN
 IF list = NIL THEN  RETURN NIL;  END;
 RETURN list^.start;
END FirstLEntry;

PROCEDURE LastLEntry (list: LIST): ENTRY;
BEGIN
 IF list = NIL THEN  RETURN NIL;  END;
 RETURN list^.end;
END LastLEntry;

PROCEDURE NextLEntry (entry: ENTRY): ENTRY;
BEGIN
 IF entry = NIL THEN  RETURN NIL;  END;
 RETURN entry^.next;
END NextLEntry;

PROCEDURE PrevLEntry (entry: ENTRY): ENTRY;
BEGIN
 IF entry = NIL THEN  RETURN NIL;  END;
 RETURN entry^.last;
END PrevLEntry;

PROCEDURE GetLEntry (entry: ENTRY; info: ADDRESS): BOOLEAN;
BEGIN
 IF (entry = NIL) OR (info = NIL) THEN  RETURN FALSE;  END;
 Copy (entry^.addr, info, entry^.size);
 RETURN TRUE;
END GetLEntry;

END M_Lists.

